theory valuePart=Main:
types boolPairs="bool \<times> bool"
constdefs Top::"boolPairs" 
         "Top==(False,False)"
constdefs T::"boolPairs"
         "T==(True,False)"
constdefs F::"boolPairs"
         "F==(False,True)"
constdefs X::"boolPairs"
         "X==(True,True)"
constdefs lub::"(boolPairs) \<Rightarrow>  (boolPairs) \<Rightarrow> (boolPairs)" 
"lub a b ==(fst a&fst b,snd a&snd b)"

constdefs OR::"boolPairs \<Rightarrow>boolPairs \<Rightarrow>boolPairs"
"OR x y==(( (fst x ) | (fst y)),((snd x ) & snd y))    "     
constdefs AND::"boolPairs \<Rightarrow>boolPairs \<Rightarrow>boolPairs"
"AND x y==(( (fst x ) & (fst y)),((snd x ) | snd y))    "   
constdefs NOT::"boolPairs \<Rightarrow>boolPairs"
"NOT x==((snd x), fst x)"

types node=nat
types time=nat
types state=" node\<Rightarrow> boolPairs"

types stateSeq= "nat\<Rightarrow>state "




constdefs stateLub::"state \<Rightarrow>state \<Rightarrow>state"
"stateLub s1 s2==\<lambda> n. (lub (s1 n) (s2 n)) "


constdefs stateSeqLub::"stateSeq\<Rightarrow>stateSeq\<Rightarrow>stateSeq"
"stateSeqLub seq1 seq2 ==
  \<lambda> t. stateLub (seq1 t) (seq2 t)"

constdefs leq::"boolPairs \<Rightarrow> boolPairs \<Rightarrow>bool" (infix "\<sqsubseteq> " 80)
"leq a b==(b=lub a b)"


constdefs stateLeq::"state \<Rightarrow>state \<Rightarrow>bool" (infix "\<sqsubseteq>\<^sub>s" 80)
"stateLeq s1 s2 == ALL n.( (s1 n)\<sqsubseteq>  (s2 n))"

constdefs stateSqLeq::"stateSeq \<Rightarrow>stateSeq \<Rightarrow>bool" (infix "\<sqsubseteq>\<^sub>s\<^sub>q " 80)
"stateSqLeq sq1 sq2==ALL t. (sq1) t \<sqsubseteq>\<^sub>s  (sq2) t"

constdefs suffix::"time \<Rightarrow> stateSeq \<Rightarrow> stateSeq"
"suffix i sq== \<lambda> t.( sq (t+i) )"


constdefs drop ::"bool \<Rightarrow> boolPairs"
"drop x==if (x=True) then T else F"

constdefs stateDrop ::"  (node \<Rightarrow> bool) \<Rightarrow>state "
"stateDrop s== \<lambda> n. drop (s n)"


constdefs stateSqDrop ::"  (time\<Rightarrow>node \<Rightarrow> bool) \<Rightarrow>stateSeq"
"stateSqDrop sq== \<lambda> t. stateDrop (sq t)"



datatype LIT=ONE|ZERO|DontCare
(*types literal="LIT\<times> node"*)
types LINE="LIT list"
types  PLA="LINE list"

constdefs funOfLit::" boolPairs \<times> LIT\<Rightarrow>boolPairs"
 "funOfLit x == if (snd x)=ONE then (fst x)
                        else if (snd x)=ZERO then (NOT (fst x))
                        else T"

consts funOfLine::" boolPairs list \<Rightarrow> LINE \<Rightarrow>boolPairs"
primrec 
  "funOfLine  bps []=T"
  "funOfLine  (bps) (el0#ls)= AND (funOfLit ((hd bps),el0)) (funOfLine  (tl bps) (ls))
  "
  
consts funOfTab::"PLA \<Rightarrow> boolPairs list \<Rightarrow> boolPairs"
primrec
  "funOfTab [] bps= F"
  "funOfTab (l#pla) bps=OR (funOfLine bps l) (funOfTab (pla) bps)"


constdefs isFullAndLine::"LINE \<Rightarrow>bool"
  "isFullAndLine line0 == ALL l. l mem line0 \<longrightarrow> l=ONE"


constdefs isAndTab::"PLA \<Rightarrow>bool"
  "isAndTab tab == length tab=1 & isFullAndLine (hd tab)"




consts PosValOfLine::" LINE  \<Rightarrow>boolPairs list"
primrec
PosVal1:  " PosValOfLine  [] =([])"

PosVal2:  " PosValOfLine  (lit#line)=
   (let otherAss=PosValOfLine  ( line) in
   (case (lit) of ONE \<Rightarrow> (T)#otherAss|
                     ZERO  \<Rightarrow> (F)#otherAss|
                     DontCare  \<Rightarrow> X#otherAss))"




 


lemma tl_map:"(tl (map s1 inps)) =map s1 (tl inps)"
apply (induct_tac inps)
apply simp
apply simp
done

lemma andMono:"\<lbrakk>a1 \<sqsubseteq> a2; b1\<sqsubseteq> b2\<rbrakk> \<Longrightarrow> (AND a1 b1)\<sqsubseteq> (AND a2 b2)"
apply -
apply(unfold AND_def leq_def)
apply(unfold lub_def)
apply(simp add:Pair_fst_snd_eq)
apply auto 
done

lemma andF:"F \<sqsubseteq> v \<Longrightarrow>  F\<sqsubseteq>(AND  v  b)" 
apply(unfold leq_def lub_def AND_def F_def,simp add:Pair_fst_snd_eq)

done

lemma andT: "\<lbrakk>T \<sqsubseteq> v1; T \<sqsubseteq> v2\<rbrakk> \<Longrightarrow>  T\<sqsubseteq>(AND  v1  v2)"
apply(unfold leq_def lub_def AND_def T_def,simp add:Pair_fst_snd_eq)
done
lemma andComm:"AND a b=AND b a" 
apply(unfold leq_def lub_def AND_def F_def,simp add:Pair_fst_snd_eq)
apply blast
done

lemma orMono:"\<lbrakk>a1 \<sqsubseteq> a2; b1\<sqsubseteq> b2\<rbrakk> \<Longrightarrow> (OR a1 b1)\<sqsubseteq> (OR a2 b2)"
apply -
apply(unfold OR_def leq_def)
apply(unfold lub_def)
apply(simp add:Pair_fst_snd_eq)
apply auto 
done

lemma orT:"\<lbrakk>T \<sqsubseteq> v1\<rbrakk>\<Longrightarrow>T \<sqsubseteq> (OR  v1 v2)"
apply (unfold OR_def leq_def)
apply(unfold lub_def T_def)
apply(simp add:Pair_fst_snd_eq)
done

lemma orF:"\<lbrakk>F \<sqsubseteq> v1;F \<sqsubseteq> v2\<rbrakk>\<Longrightarrow>  F\<sqsubseteq> (OR v1 v2)"
apply (unfold OR_def leq_def)
apply(unfold lub_def F_def)
apply(simp add:Pair_fst_snd_eq)
done

lemma orComm:"OR a b=OR b a" 
apply(unfold leq_def lub_def OR_def F_def,simp add:Pair_fst_snd_eq)
apply blast
done
lemma notMono:"\<lbrakk>a1 \<sqsubseteq> a2\<rbrakk> \<Longrightarrow> (NOT a1 )\<sqsubseteq> (NOT a2 )"
  apply(unfold NOT_def leq_def)
  apply(unfold lub_def)
  apply(simp add:Pair_fst_snd_eq)  
  done

lemma lubMono:"\<lbrakk>a1 \<sqsubseteq> a2; b1\<sqsubseteq> b2\<rbrakk> \<Longrightarrow> (lub a1 b1)\<sqsubseteq> (lub a2 b2)"
  apply(unfold NOT_def leq_def)
  apply(unfold lub_def)
  apply(simp add:Pair_fst_snd_eq)  
  apply auto
  done
  
lemma lubSelf: " a \<sqsubseteq> (lub a  b )"
  apply(unfold leq_def lub_def)
  apply auto
done

lemma lubComm:" lub a b= lub b a"
  apply(unfold leq_def lub_def)
  apply auto
done


lemma lubSelf: " a \<sqsubseteq> (lub a  b )"
  apply(unfold leq_def lub_def)
  apply auto
done

lemma lubId:" lub a a=a"
  by(unfold  lub_def,  auto)

lemma lubComm:" lub a b= lub b a"
  apply(unfold leq_def lub_def)
  apply auto
done



lemma XIsLeastValue:
  " X \<sqsubseteq> a"
  apply(unfold leq_def)
  apply(unfold lub_def)
  apply(unfold X_def,simp)
done

lemma lubMeans:
  "\<lbrakk> x \<sqsubseteq> a; y\<sqsubseteq> a\<rbrakk> \<Longrightarrow> lub x y  \<sqsubseteq> a"
  apply(unfold leq_def lub_def)
  apply auto 
  apply(simp add:Pair_fst_snd_eq)
  done
  

lemma leqReflexive:"a \<sqsubseteq> a"
  apply(unfold leq_def)
  apply(unfold lub_def)
  apply(simp add:Pair_fst_snd_eq)  
  done


lemma leqIsTrans:
  "\<lbrakk>a \<sqsubseteq> b; b \<sqsubseteq> c\<rbrakk>\<Longrightarrow> a\<sqsubseteq> c" 
  apply(unfold leq_def lub_def) 
  apply(simp add:Pair_fst_snd_eq)
  apply auto
  done

lemma sqLeqIsTrans:
  "\<lbrakk>a \<sqsubseteq>\<^sub>s\<^sub>q  b; b \<sqsubseteq>\<^sub>s\<^sub>q  c\<rbrakk>\<Longrightarrow> a\<sqsubseteq>\<^sub>s\<^sub>q  c" 
  apply(unfold stateSqLeq_def stateLeq_def ) 
  apply(auto dest:leqIsTrans)
  apply(blast dest:leqIsTrans)
  done
lemma sqLeqIsTrans:
"\<lbrakk>sq0  \<sqsubseteq>\<^sub>s\<^sub>q sq1; sq1  \<sqsubseteq>\<^sub>s\<^sub>q sq2 \<rbrakk>\<Longrightarrow> sq0  \<sqsubseteq>\<^sub>s\<^sub>q sq2"
  apply(unfold stateSqLeq_def stateLeq_def)
  apply(rule allI)+
  apply(blast intro:leqIsTrans)
done


lemma eqIByLeq:
"\<lbrakk> a \<sqsubseteq> b; b \<sqsubseteq> a\<rbrakk> \<Longrightarrow> a=b"
apply(unfold leq_def lub_def)
apply(simp add:Pair_fst_snd_eq)
apply auto
done

lemma Leq1I:
"\<lbrakk> a=b\<rbrakk> \<Longrightarrow>a \<sqsubseteq> b"
apply(unfold leq_def lub_def)
apply(simp add:Pair_fst_snd_eq)
done

lemma Leq2I:
"\<lbrakk> a=b\<rbrakk> \<Longrightarrow>b \<sqsubseteq> a"
apply(unfold leq_def lub_def)
apply(simp add:Pair_fst_snd_eq)
done


lemma sqLeqIsReflexive:" sq \<sqsubseteq>\<^sub>s\<^sub>q sq"
  apply(unfold stateSqLeq_def  stateLeq_def)
  apply(rule allI)+
  apply(rule leqReflexive)
  done


lemma funOfLitIsMono:"a1 \<sqsubseteq>a2 \<Longrightarrow>funOfLit(a1,sign0) \<sqsubseteq>funOfLit(a2,sign0)"
  apply(case_tac sign0)
  apply(unfold funOfLit_def)
  apply simp
  apply simp
  apply(simp add:notMono) 
  apply(simp )
  apply auto
  apply(simp add:leqReflexive)
  done

lemma hdMap1:"inps~=[]\<longrightarrow>hd (map s1 inps)=s1 (hd inps) & (hd inps) mem inps"
  apply(induct_tac inps)
  apply simp
  apply simp
done

lemma hdMap:"inps~=[]\<Longrightarrow>hd (map s1 inps)=s1 (hd inps) & (hd inps) mem inps"
  apply(cut_tac hdMap1[where inps="inps" and ?s1.0="s1" ])
  apply simp
 
done

lemma funOfLineIsMono:"(s1 \<sqsubseteq>\<^sub>s s2) \<longrightarrow>
  ( ALL inps. length inps=length line\<longrightarrow>
   (funOfLine (map s1 inps)) line\<sqsubseteq>  (funOfLine (map s2 inps)) line )"
apply(induct_tac line)
apply simp
apply(unfold leq_def lub_def)
apply simp
apply(fold  lub_def leq_def)

apply simp
apply(rule impI)+
apply(rule allI,rule impI)
apply(subgoal_tac "(tl (map s1 inps)) =map s1 (tl inps) ")
apply(subgoal_tac "(tl (map s2 inps)) =map s2 (tl inps)")
prefer 2
apply(rule tl_map)
apply simp
prefer 2
apply(rule tl_map) thm  andMono
apply(rule andMono)
apply(subgoal_tac "hd (map s1 inps)=s1 (hd inps)& (hd inps) mem inps")
apply(subgoal_tac "hd (map s2 inps)=s2 (hd inps)& (hd inps) mem inps")
apply(rule funOfLitIsMono)
apply(unfold stateLeq_def)
apply simp
apply(rule hdMap)
apply(subgoal_tac "0<length inps")
apply(simp (no_asm_use)  add: length_greater_0_conv )
apply simp
apply(rule hdMap)
apply(subgoal_tac "0<length inps")
apply(simp (no_asm_use)  add: length_greater_0_conv )
apply simp
apply auto
done

lemma funOfLineIsMono:"\<lbrakk>(s1 \<sqsubseteq>\<^sub>s s2); length inps=length tab\<rbrakk>\<Longrightarrow>  
   (funOfLine (map s1 inps)) tab\<sqsubseteq>  (funOfLine (map s2 inps)) tab "
apply(cut_tac funOfLineIsMono[where ?s1.0="s1" and ?s2.0="s2" and line="tab"])
apply blast
done

lemma funOfLineIsMonoAny:"
  ( ALL inps. length inps=length line\<longrightarrow>(ALL n. n mem inps \<longrightarrow> s1 n \<sqsubseteq> s2 n) \<longrightarrow>
   (funOfLine (map s1 inps)) line\<sqsubseteq>  (funOfLine (map s2 inps)) line )"
apply(induct_tac line)
apply simp
apply(unfold leq_def lub_def)
apply simp
apply(fold  lub_def leq_def)

apply simp
apply(rule allI,(rule impI)+)
apply(subgoal_tac "(tl (map s1 inps)) =map s1 (tl inps)")
apply(subgoal_tac "(tl (map s2 inps)) =map s2 (tl inps)")
prefer 2
apply(rule tl_map)
apply simp
prefer 2
apply(rule tl_map) thm  andMono
apply(rule andMono)
apply(subgoal_tac "hd (map s1 inps)=s1 (hd inps)& (hd inps) mem inps")
apply(subgoal_tac "hd (map s2 inps)=s2 (hd inps)& (hd inps) mem inps")
apply(rule funOfLitIsMono)
apply simp
apply(rule hdMap)
apply(subgoal_tac "0<length inps")
apply(simp (no_asm_use)  add: length_greater_0_conv )
apply simp
apply(rule hdMap)
apply(subgoal_tac "0<length inps")
apply(simp (no_asm_use)  add: length_greater_0_conv )
apply simp
apply(drule_tac x="(tl inps)" in spec)
apply simp
apply(subgoal_tac "(\<forall>n. n mem tl inps \<longrightarrow> s1 n \<sqsubseteq>  s2 n)")
apply blast
apply(rule allI)
apply(rule impI)
apply(subgoal_tac "n mem inps")
apply blast
apply(case_tac inps)
apply auto
done

lemma funOfLineIsMonoAny:"\<lbrakk>(ALL n. n mem inps \<longrightarrow> s1 n \<sqsubseteq> s2 n); length inps=length tab\<rbrakk>\<Longrightarrow>  
   (funOfLine (map s1 inps)) tab\<sqsubseteq>  (funOfLine (map s2 inps)) tab "
apply(cut_tac funOfLineIsMonoAny[where ?s1.0="s1" and ?s2.0="s2" and line="tab"])
apply blast
done

lemma funOfTabIsMono:
shows "(s1 \<sqsubseteq>\<^sub>s s2) \<longrightarrow>(ALL l. (l:set tab)\<longrightarrow> length inps=length l) \<longrightarrow>
  funOfTab tab (map s1 inps) \<sqsubseteq>  funOfTab tab (map s2 inps)"
apply -
apply(induct_tac tab)
apply simp
apply((rule impI)+,rule leqReflexive)
apply simp
apply((rule impI)+,rule orMono)
apply(rule funOfLineIsMono)
apply blast+
done

lemma funOfTabIsMono:
shows "\<lbrakk>(s1 \<sqsubseteq>\<^sub>s s2) ;(ALL l. (l:set tab)\<longrightarrow> length inps=length l) \<rbrakk>\<Longrightarrow>
  funOfTab tab (map s1 inps) \<sqsubseteq>  funOfTab tab (map s2 inps)"
apply(cut_tac  funOfTabIsMono[where ?s1.0="s1" and ?s2.0="s2" and tab="tab"])
apply blast
done

lemma funOfTabIsMonoAny:
shows "(ALL n. n mem inps \<longrightarrow> s1 n \<sqsubseteq> s2 n) \<longrightarrow>(ALL l. (l:set tab)\<longrightarrow> length inps=length l) \<longrightarrow>
  funOfTab tab (map s1 inps) \<sqsubseteq>  funOfTab tab (map s2 inps)"
apply -
apply(induct_tac tab)
apply simp
apply((rule impI)+,rule leqReflexive)
apply simp
apply((rule impI)+,rule orMono)
apply(rule funOfLineIsMonoAny)
apply blast+
done

lemma funOfTabIsMonoAny:
shows "\<lbrakk>(ALL n. n mem inps \<longrightarrow> s1 n \<sqsubseteq> s2 n);(ALL l. (l:set tab)\<longrightarrow> length inps=length l) \<rbrakk>\<Longrightarrow>
  funOfTab tab (map s1 inps) \<sqsubseteq>  funOfTab tab (map s2 inps)"
apply(cut_tac  funOfTabIsMonoAny[where ?s1.0="s1" and ?s2.0="s2" and tab="tab"])
apply blast
done

consts stateSqLeq1::"boolPairs list \<Rightarrow> boolPairs list  \<Rightarrow> bool"
primrec
stateSqLeq1Nil: "stateSqLeq1 [] ls= True"
stateSqLeq1Cons:"stateSqLeq1 (x#xs) ys = ((x \<sqsubseteq> (hd ys)) & (stateSqLeq1 xs (tl ys)))"

lemma hdStateSqLeq1:
"\<lbrakk>stateSqLeq1 stateLs1 stateLs2; length stateLs1 = Suc (length list)\<rbrakk>
       \<Longrightarrow> hd stateLs1 \<sqsubseteq>  hd stateLs2"
apply(simp (no_asm_use)  only:length_Suc_conv )
apply(erule exE)+
apply(simp add:stateSqLeq1Cons)
done

lemma funOfLineIsMono:
"ALL stateLs1 stateLs2. stateSqLeq1 stateLs1   stateLs2 \<longrightarrow>
  ( length  stateLs1=length l\<longrightarrow>(funOfLine stateLs1 l) \<sqsubseteq>  (funOfLine stateLs2 l))"
apply(induct_tac l)
apply simp
apply(unfold leq_def lub_def)
apply simp
apply(fold  lub_def leq_def)

apply simp
apply((rule allI)+,(rule impI)+)
apply(rule andMono)
apply(rule funOfLitIsMono)
apply(rule hdStateSqLeq1)
apply assumption+ 
apply(drule_tac x="tl stateLs1" in spec)
apply(drule_tac x="tl stateLs2" in spec)
apply(simp (no_asm_use)  only:length_Suc_conv )
apply(erule exE)+
apply(simp add:stateSqLeq1Cons)
done

lemma funOfLineIsMono:
"\<lbrakk> stateSqLeq1 stateLs1   stateLs2 ;
  length  stateLs1=length l\<rbrakk> \<Longrightarrow>((funOfLine stateLs1 l) \<sqsubseteq>  (funOfLine stateLs2 l))"
apply(cut_tac funOfLineIsMono[where l="l"])
apply blast
done

lemma andLinePropF:
assumes a:"F \<sqsubseteq> v"
shows 
"ALL bps.  v mem bps \<longrightarrow> length bps=length line \<longrightarrow> isFullAndLine line\<longrightarrow>F \<sqsubseteq> funOfLine bps line"
  (is "?P line")
proof(induct_tac line)
  show "?P []"
    apply(rule allI,rule impI,rule impI)
    apply simp
    done
next 
  fix a list
  assume a0:"?P list"
  show "?P (a # list)"
    proof(rule allI,(rule impI)+)
      fix bps
      assume a1:"v mem bps" and
      a2:"length bps = length (a # list)" and
      a3:"isFullAndLine (a # list)"
      show "F \<sqsubseteq>funOfLine bps (a # list)"
      proof(simp)
    let ?P=
     "  F \<sqsubseteq>AND (funOfLit (hd bps, a)) (funOfLine (tl bps) list) "
    from a2 have b1:"EX n. length bps=Suc n"
      apply auto
      done
    then obtain n where b1:"length bps=Suc n"
      by auto
    then have b2:"EX y ys.( bps = y # ys & length ys = n)"
      by (simp add:length_Suc_conv)
    then  obtain y ys  where b2:"bps = y # ys "
      by auto
    from a1 b2 have "y=v | v mem ys"
      apply -
      apply(case_tac "y=v", auto)
      done
    moreover
    {assume b3:"y=v"
      have b4:"a=ONE"
        apply(cut_tac a3,unfold isFullAndLine_def)
        apply auto
        done
      have " F \<sqsubseteq>(funOfLit (hd bps, a))"
        apply(cut_tac  b3 b2 b4 ,simp add:funOfLit_def)
	apply(cut_tac a,simp)
        done
      then have ?P
        apply - 
	apply(rule_tac andF) 
        apply (simp)
        done
    }
    moreover
    {assume b3:" v mem ys"
      have b4: "length ys = length list "
        apply(cut_tac a2 b2)
        by auto

      have "isFullAndLine list "
        apply(cut_tac a3)
        apply(unfold isFullAndLine_def)
        by auto
      
      from b3 a0 this b4  have "F \<sqsubseteq> funOfLine ys list "
        by auto
      from b2 this have "F \<sqsubseteq>funOfLine (tl bps) list "
        by auto
      then have ?P
        apply -
	apply(subgoal_tac "AND (funOfLit (hd bps, a)) (funOfLine (tl bps) list)=
	  AND (funOfLine (tl bps) list) (funOfLit (hd bps, a))")
	apply simp
	apply(rule_tac andF) 
        apply (simp)
        apply (rule andComm)
        done
      }
      ultimately show ?P by blast
    qed
      qed
    qed

lemma andLinePropF:
"\<lbrakk>F \<sqsubseteq> v; v mem bps ; length bps=length line; isFullAndLine line\<rbrakk>\<Longrightarrow> F \<sqsubseteq> funOfLine bps line"
  apply(cut_tac  andLinePropF[where line="line"])
  apply blast
  apply simp
  done



(*a n usual line, need not be a fullAndLine*)
lemma aLinePropT:
shows 
"ALL bps.   length bps=length line \<longrightarrow> bps= PosValOfLine line\<longrightarrow>T\<sqsubseteq> funOfLine bps line"
  (is "?P line")
proof(induct_tac line)
  show "?P []"
    apply(rule allI,rule impI,rule impI)
    apply simp
    apply(rule leqReflexive)
    done
next 
  fix a list
  assume a0:"?P list"
  show "?P ((a::LIT) # list)"
    proof(rule allI,(rule impI)+)
      fix bps
      assume a2:"length bps = length (a # list)" and
	a3:"bps = PosValOfLine  (a # list)" 
      show "T \<sqsubseteq> funOfLine bps (a # list) "
      proof(simp)
	let ?P=
	  " T \<sqsubseteq> AND (funOfLit (hd bps, a)) (funOfLine (tl bps) list) "
	from a2 have b1:"EX n. length bps=Suc n"
	  apply auto
	  done
	then obtain n where b1:"length bps=Suc n"
	  by auto
	then have b2:"EX y ys.( bps = y # ys & length ys = n)"
	  by (simp add:length_Suc_conv)
	then  obtain y ys  where b2:"bps = y # ys "
	  by auto
	have b3:" T \<sqsubseteq>(funOfLit (hd bps, (a))) &  ys = PosValOfLine ( list)"
	  apply(cut_tac a3 b2  ,simp add:funOfLit_def)   
	  apply(case_tac "a::LIT", simp add:Let_def)
	  apply(rule leqReflexive)
	  apply( simp add:Let_def)
	  apply(unfold T_def F_def leq_def lub_def NOT_def,simp add:Pair_fst_snd_eq)
	  apply (simp add:Let_def)
	  done
	have b4a: "length ys = length list "
	  apply(cut_tac a2 b2)
	  by auto

	
      
	from b3 a0 this b4a  have "T \<sqsubseteq>funOfLine ys list "
	  by auto
	from b2 this have "T \<sqsubseteq>funOfLine (tl bps) list"
          by auto
	from b3 this show ?P
	  apply -
	  by (rule andT,auto)
      
      qed
    qed
  qed

lemma aLinePropT:
shows 
"\<lbrakk> length bps=length line ; bps= PosValOfLine line\<rbrakk>\<Longrightarrow>T\<sqsubseteq> funOfLine bps line"
proof(cut_tac  aLinePropT[where line="line"],blast) qed

consts listLeq::"boolPairs list \<Rightarrow> boolPairs list \<Rightarrow> bool"
primrec 
  listLeq1:"listLeq [] l2=True"
  listLeq2:"listLeq (a#list0) l =(case l of []\<Rightarrow> False |
                            (b#list1)\<Rightarrow> (a \<sqsubseteq> b)& (listLeq list0 list1))"
lemma leqIsTrans:
  "\<lbrakk>a \<sqsubseteq> b; b \<sqsubseteq> c\<rbrakk>\<Longrightarrow> a\<sqsubseteq> c" 
  apply(unfold leq_def lub_def) 
  apply(simp add:Pair_fst_snd_eq)
  apply auto
  done

lemma listLeqTrans:" ALL list2 list3. listLeq list1 list2 \<longrightarrow> listLeq list2 list3\<longrightarrow>listLeq list1 list3"
  apply(induct_tac list1)
  apply simp
  apply(rule allI)
  apply(induct_tac list2)
  apply simp
  apply(rule allI)
  apply(induct_tac list3)
  apply simp
  apply(rule impI)+
  apply( simp (no_asm))
  apply(simp)
  apply(erule conjE)+
  apply(rule conjI)
  apply(rule_tac b="aa" in  leqIsTrans)
  apply simp+
  done

lemma listLeqTrans:"
\<lbrakk>listLeq list1 list2 ; listLeq list2 list3\<rbrakk>\<Longrightarrow>listLeq list1 list3"
apply(cut_tac  listLeqTrans[where ?list1.0="list1"])
apply blast
done







lemma aLinePropT1:
shows 
"ALL bps.   length bps=length line \<longrightarrow> listLeq (PosValOfLine line) bps\<longrightarrow>T\<sqsubseteq> funOfLine bps line"
  (is "?P line")
proof(induct_tac line)
  show "?P []"
    apply(rule allI,rule impI,rule impI)
    apply simp
    apply(rule leqReflexive)
    done
next 
  fix a list
  assume a0:"?P list"
  show "?P ((a::LIT) # list)"
    proof(rule allI,(rule impI)+)
      fix bps
      assume a2:"length bps = length (a # list)" and
	a3:"listLeq (PosValOfLine (a # list)) bps" 
      show "T \<sqsubseteq> funOfLine bps (a # list) "
      proof(simp)
	let ?P=
	  " T \<sqsubseteq> AND (funOfLit (hd bps, a)) (funOfLine (tl bps) list) "
	from a2 have b1:"EX n. length bps=Suc n"
	  apply auto
	  done
	then obtain n where b1:"length bps=Suc n"
	  by auto
	then have b2:"EX y ys.( bps = y # ys & length ys = n)"
	  by (simp add:length_Suc_conv)
	then  obtain y ys  where b2:"bps = y # ys "
	  by auto
	have b3:" T \<sqsubseteq>(funOfLit (hd bps, (a))) & listLeq (PosValOfLine list) ys"
	  apply(cut_tac a3 b2  ,simp add:funOfLit_def)   
	  apply(case_tac "a::LIT", simp add:Let_def)
	  apply( simp add:Let_def)
	  apply(unfold T_def F_def leq_def lub_def NOT_def,simp add:Pair_fst_snd_eq)
	  apply (simp add:Let_def)
	  done
	have b4a: "length ys = length list "
	  apply(cut_tac a2 b2)
	  by auto

	
      
	from b3 a0 this b4a  have "T \<sqsubseteq>funOfLine ys list "
	  by auto
	from b2 this have "T \<sqsubseteq>funOfLine (tl bps) list"
          by auto
	from b3 this show ?P
	  apply -
	  by (rule andT,auto)
      
      qed
    qed
  qed



lemma aLinePropT1:
shows 
"\<lbrakk> length bps=length line ;listLeq (PosValOfLine line) bps\<rbrakk>\<Longrightarrow>T\<sqsubseteq> funOfLine bps line"
proof(cut_tac  aLinePropT1[where line="line"],blast) qed



lemma andLinePropT: 
  
  "ALL bps.  (ALL bv. bv mem bps \<longrightarrow> T \<sqsubseteq>bv ) \<longrightarrow>
  length bps=length line \<longrightarrow> isFullAndLine line\<longrightarrow>  T \<sqsubseteq> funOfLine bps line"
  (is "?P line")
proof(induct_tac line)
  show "?P []"
    apply(rule allI,rule impI,rule impI)
    apply simp
    apply(rule impI, rule leqReflexive)
    done
next 
  fix a list
  assume a0:"?P list"
  show "?P (a # list)"
    proof(rule allI,(rule impI)+)
      fix bps
      assume a1:"(\<forall>bv. bv mem bps \<longrightarrow> T \<sqsubseteq>bv )" and
      a2:"length bps = length (a # list)" and
      a3:"isFullAndLine (a # list)"
      show "T \<sqsubseteq> funOfLine bps (a # list) "
      proof(simp)
    let ?P=
     " T \<sqsubseteq> AND (funOfLit (hd bps, a)) (funOfLine (tl bps) list) "
    from a2 have b1:"EX n. length bps=Suc n"
      apply auto
      done
    then obtain n where b1:"length bps=Suc n"
      by auto
    then have b2:"EX y ys.( bps = y # ys & length ys = n)"
      by (simp add:length_Suc_conv)
    then  obtain y ys  where b2:"bps = y # ys "
      by auto
    from a1 b2 have b3:"T \<sqsubseteq>y &( ALL y. y mem ys \<longrightarrow>T \<sqsubseteq>y)"
      apply -
      apply (simp del:pair_collapse split_paired_All)     
      done
    

    have b4:"a=ONE"
      apply(cut_tac a3,unfold isFullAndLine_def)
      apply auto
      done
    have b5:"T \<sqsubseteq>(funOfLit (hd bps, a))"
      apply(cut_tac b3 b2 b4 ,simp add:funOfLit_def)
      done
    
    have b4a: "length ys = length list "
      apply(cut_tac a2 b2)
      by auto

    have b4b:"isFullAndLine list "
      apply(cut_tac a3)
      apply(unfold isFullAndLine_def)
      by auto
      
    from b3 a0 this b4a  have "T \<sqsubseteq>funOfLine ys list "
      by auto
    from b2 this have "T \<sqsubseteq>funOfLine (tl bps) list"
        by auto
    from b5 this show ?P
      apply -
      by (rule andT)
      
      qed
    qed
  qed

lemma andLinePropT:
  "\<lbrakk>(ALL bv. bv mem bps \<longrightarrow> T \<sqsubseteq>bv) ;
  length bps=length line; isFullAndLine line\<rbrakk>\<Longrightarrow> T \<sqsubseteq> funOfLine bps line"
  apply(cut_tac  andLinePropT[where line="line"])
  apply blast
  done

lemma andTabPropF:
"\<lbrakk>F \<sqsubseteq>v; v mem bps ; length bps=length (hd tab) ; isAndTab tab\<rbrakk>\<Longrightarrow>F \<sqsubseteq> funOfTab tab bps"
apply(unfold isAndTab_def)
apply simp
apply(erule conjE)
apply(subgoal_tac "EX y ys.( tab = y # ys & length ys = 0)")
apply(erule exE)+
apply(erule conjE)
apply simp
apply(subgoal_tac "F \<sqsubseteq>funOfLine bps y")
apply(simp add:OR_def F_def)
apply(rule andLinePropF)
apply assumption+
apply(simp add:length_Suc_conv)
done
lemma orT:"\<lbrakk>T \<sqsubseteq>a\<rbrakk> \<Longrightarrow>T \<sqsubseteq> OR  a b "
apply(unfold leq_def lub_def OR_def T_def,simp add:Pair_fst_snd_eq)
done
lemma andTabPropT:
"\<lbrakk>ALL bv. bv mem bps \<longrightarrow>  T \<sqsubseteq>bv ; length bps=length (hd tab) ; isAndTab tab\<rbrakk>\<Longrightarrow>  T \<sqsubseteq>funOfTab tab bps "
apply(unfold isAndTab_def)
apply( simp del:pair_collapse split_paired_All)
apply(erule conjE)
apply(subgoal_tac "EX y ys.( tab = y # ys & length ys = 0)")
apply(erule exE)+
apply(erule conjE)
apply(simp del:pair_collapse split_paired_All)
apply(subgoal_tac " T \<sqsubseteq>funOfLine bps y")
apply(rule orT)
apply assumption
apply(rule andLinePropT)
apply assumption+
apply(simp add:length_Suc_conv)
done

lemma orTabPropT:
shows "ALL bps. line mem tab\<longrightarrow>length bps=length line \<longrightarrow>
  bps= PosValOfLine line\<longrightarrow> T\<sqsubseteq> funOfTab tab bps " (is "?P tab")
proof(induct_tac tab)
  show "?P []"
  proof(simp) qed
next
  fix a list
  assume a0:"?P list"
  show "?P (a # list)"
  proof(rule allI, (rule impI)+)
    fix bps
    assume a1:"line mem a # list" and
      a2:"length bps = length line" and
      a3:" bps = PosValOfLine line"
    show "T \<sqsubseteq> funOfTab (a # list) bps"
    proof -
      have "a=line | line mem list" 
	apply(cut_tac a1,simp)
	by(case_tac "a=line",auto)
      moreover
      {assume b1:"a=line"
	have b2:"T\<sqsubseteq> funOfLine bps line"
	proof(rule       aLinePropT) qed
	have "T \<sqsubseteq>  funOfTab (a # list) bps"
	proof(simp,rule_tac orT,cut_tac b1 b2,auto) qed
      }
      moreover
      {assume b1:"line mem list"
	from a0 have b2:" T\<sqsubseteq> funOfTab list bps "
	  apply -
	  apply(drule_tac x="bps" in spec)
	  apply(cut_tac a2 a3 b1)
	  by blast
	then have "T \<sqsubseteq>  funOfTab (a # list) bps"
	proof(simp) 
	  have c1:"OR (funOfLine bps a) (funOfTab list bps)=
	    OR  (funOfTab list bps) (funOfLine bps a)"
	    by (rule_tac orComm)
	  have c2:"T \<sqsubseteq> OR  (funOfTab list bps) (funOfLine bps a)"
	  proof(rule_tac orT,cut_tac  b2,auto) qed
	  with c1 show "T \<sqsubseteq>OR (funOfLine bps a) (funOfTab list bps)"
	    by auto
	qed
      }
      ultimately   show "T \<sqsubseteq> funOfTab (a # list) bps"
	by blast
    qed
  qed
qed
      

lemma orTabPropT:
 "\<lbrakk> line mem tab;length bps=length line ;
  bps= PosValOfLine line\<rbrakk> \<Longrightarrow>  T\<sqsubseteq> funOfTab tab bps "
apply(cut_tac 	orTabPropT[where tab="tab" and line="line"],blast)
done
 lemma orTabPropT1:
shows "ALL bps. line mem tab\<longrightarrow>length bps=length line \<longrightarrow>
   listLeq (PosValOfLine line) bps\<longrightarrow> T\<sqsubseteq> funOfTab tab bps " (is "?P tab")
proof(induct_tac tab)
  show "?P []"
  proof(simp) qed
next
  fix a list
  assume a0:"?P list"
  show "?P (a # list)"
  proof(rule allI, (rule impI)+)
    fix bps
    assume a1:"line mem a # list" and
      a2:"length bps = length line" and
      a3:" listLeq (PosValOfLine line) bps"
    show "T \<sqsubseteq> funOfTab (a # list) bps"
    proof -
      have "a=line | line mem list" 
	apply(cut_tac a1,simp)
	by(case_tac "a=line",auto)
      moreover
      {assume b1:"a=line"
	have b2:"T\<sqsubseteq> funOfLine bps line"
	proof(rule       aLinePropT1) qed
	have "T \<sqsubseteq>  funOfTab (a # list) bps"
	proof(simp,rule_tac orT,cut_tac b1 b2,auto) qed
      }
      moreover
      {assume b1:"line mem list"
	from a0 have b2:" T\<sqsubseteq> funOfTab list bps "
	  apply -
	  apply(drule_tac x="bps" in spec)
	  apply(cut_tac a2 a3 b1)
	  by blast
	then have "T \<sqsubseteq>  funOfTab (a # list) bps"
	proof(simp) 
	  have c1:"OR (funOfLine bps a) (funOfTab list bps)=
	    OR  (funOfTab list bps) (funOfLine bps a)"
	    by (rule_tac orComm)
	  have c2:"T \<sqsubseteq> OR  (funOfTab list bps) (funOfLine bps a)"
	  proof(rule_tac orT,cut_tac  b2,auto) qed
	  with c1 show "T \<sqsubseteq>OR (funOfLine bps a) (funOfTab list bps)"
	    by auto
	qed
      }
      ultimately   show "T \<sqsubseteq> funOfTab (a # list) bps"
	by blast
    qed
  qed
qed
      

lemma orTabPropT:
 "\<lbrakk> line mem tab;length bps=length line ;
 listLeq (PosValOfLine line) bps\<rbrakk> \<Longrightarrow>  T\<sqsubseteq> funOfTab tab bps "
apply(cut_tac 	orTabPropT1[where tab="tab" and line="line"],blast)
done

types LITValuePair="LIT\<times>  boolPairs"

constdefs contradict:: "LITValuePair \<Rightarrow> bool"
"contradict pair \<equiv>  (fst pair)=ONE &   F\<sqsubseteq>(snd pair) |
 (fst pair)=ZERO & T\<sqsubseteq>(snd pair)"

lemma aLinePropF:
shows
"ALL bps.   length bps=length line \<longrightarrow> 
  (EX asOfVal. asOfVal mem ( zip line bps) & contradict asOfVal)\<longrightarrow>F \<sqsubseteq> funOfLine bps line"
  (is "?P line")
proof(induct_tac line)
  show "?P []" by auto
next
  fix a line
  assume b1:"?P line"
  show "?P (a#line)"
  proof(rule allI, (rule impI)+)
    fix bps
    assume c1:"length bps = length (a # line)" and
    c2:"\<exists>asOfVal. asOfVal mem zip (a # line) bps \<and> contradict asOfVal" 
    show " F \<sqsubseteq>  funOfLine bps (a # line)"
    proof -
      have " length bps =Suc (length line) "
	proof(cut_tac c1, auto) qed
	then have d2:"EX y ys.( bps = y # ys & length ys = length line)"
	  by (simp add:length_Suc_conv)
	then  obtain z zs  where d2:"bps = z # zs& length zs = length line"
	  by auto

      from c2 obtain asOfVal where 
	d1:"asOfVal mem zip (a # line) bps \<and> contradict asOfVal" 
	by blast
      
      with d2 have "asOfVal=(a,z)| asOfVal mem zip ( line) zs" 
	apply -
	apply simp
	by(case_tac "(a, z) = asOfVal",auto)
      moreover
      {assume e1:"asOfVal=(a,z)"
	with d1 d2 have ?thesis
	proof(simp add:contradict_def funOfLit_def,case_tac "a::LIT",auto,
              rule_tac andF,simp, rule_tac andF,
	      simp add:F_def T_def NOT_def leq_def lub_def,
	      simp add:Pair_fst_snd_eq) 
	qed
      } 
      moreover
      {assume e1:"asOfVal mem zip line zs"
	with d1 d2 b1 have e2: "F \<sqsubseteq> funOfLine zs line"
	  by(drule_tac x="zs" in spec,blast)
	with d2  have ?thesis
	proof(simp)
	  have " AND (funOfLit (z, a)) (funOfLine zs line)=
	     AND  (funOfLine zs line)(funOfLit (z, a))"
	    by (rule andComm)
	  with  e2 show   "F \<sqsubseteq>AND (funOfLit (z, a)) (funOfLine zs line)"
	    apply - apply(simp)
	    apply(rule_tac andF)
	    by simp
	qed
      }
      ultimately show ?thesis by blast
      qed
    qed
  qed

lemma aLinePropF:
  assumes a1:"length bps=length line" and
  a2:"(EX asOfVal. asOfVal mem ( zip line bps) & contradict asOfVal)"
  shows
  "F \<sqsubseteq> funOfLine bps line"
proof(cut_tac aLinePropF[where line="line"] a1 a2,blast) qed

lemma orTabPropF1:
   
  shows "ALL bps. ( (\<forall> l. (l\<in> (set tab)) \<longrightarrow> length l=length bps)\<longrightarrow>
  (ALL line. line mem  tab \<longrightarrow> (EX asOfVal. asOfVal mem ( zip line bps) & contradict asOfVal))  \<longrightarrow>
  F \<sqsubseteq> funOfTab tab bps)" (is "?P tab")
proof(induct_tac tab)
  show "?P []" by ( simp,rule leqReflexive)
next
  fix a tab
  assume b1:"?P tab"
  show "?P (a#tab)"
  proof(rule allI,(rule impI)+)
    fix bps
    assume c1:"\<forall>l. l \<in> set (a # tab) \<longrightarrow> length l = length bps" and
    c2:"\<forall>line. line mem a # tab \<longrightarrow> (\<exists>asOfVal. asOfVal mem zip line (bps:: boolPairs list)
      \<and> contradict asOfVal)"
    show " F \<sqsubseteq>  funOfTab (a # tab) bps"
    proof(simp)
      from c2 have c3:"(\<exists>asOfVal. asOfVal mem zip a (bps:: boolPairs list)
      \<and> contradict asOfVal)"
      proof(drule_tac x="a" in spec,simp) qed
      from c2 
      have c4:"\<forall>line. line mem  tab \<longrightarrow> (\<exists>asOfVal. asOfVal mem zip line (bps:: boolPairs list)
	\<and> contradict asOfVal)"
      proof(rule_tac allI,rule_tac impI,force) qed
      from c1 have c5:"length a=length bps" by simp
      from c1 have c6:"\<forall> l. l\<in> (set tab) \<longrightarrow> length l=length bps"
	by simp
      from c3 c5 have c7:"F \<sqsubseteq>(funOfLine bps a)" 
	by(rule_tac  aLinePropF,auto)
      from c4 c6 b1 have "F \<sqsubseteq>  funOfTab ( tab) bps"
      proof(drule_tac x="bps" in spec, blast)qed
      with c7 show " F \<sqsubseteq>  OR (funOfLine bps a) (funOfTab tab bps)"
      proof(rule_tac orF)qed
    qed
  qed
qed
	
lemma orTabPropF1:
  assumes a1:" (\<forall> l. (l\<in> (set tab)) \<longrightarrow> length l=length bps)" and
  a2:"(ALL line. line mem  tab \<longrightarrow> (EX asOfVal. asOfVal mem ( zip line bps) & contradict asOfVal))"

  shows "  F \<sqsubseteq> funOfTab tab bps"
proof(cut_tac orTabPropF1[where tab="tab"] a1 a2, blast)qed

lemma arith1:" 0< x \<Longrightarrow> EX n. x= Suc n" by arith

consts consListAux::"'a list \<Rightarrow> node list"
primrec 
consAuxNil:" consListAux []=[]"
consAuxCons:"consListAux (x#xs)=(length xs +1) # (consListAux xs)"

lemma conAuxSameLen:" length (consListAux xs)=length xs"
  apply(induct_tac xs)
  apply auto
done


end 
